#include "stdafx.h"
#include "board.h"
#include <math.h>

namespace {
	// (x, y)ՏɂꍇAtrue Ԃ
	inline bool isOn(int x, int y)
	{
		return ((UINT)x < Board::SIZE) & ((UINT)y < Board::SIZE);
	}

	// Zp`WXN[W֕ϊ
	POINT HexPosToScreenPos(int x, int y)
	{
		static const float r = Board::CELL_SIZE / 2.0f;
		static const float r3 = sqrt(3.0f);

		POINT result;

		if (y % 2) {
//			result.x = (x * r * r3) + (r * r3);
//			result.y = y * r * 1.5f + r;
			result.x = (x + 1.0f) * (r * r3);
			result.y = (y * 1.5f + 1.0f) * r;
		} else {
//			result.x = (x * r * r3) + (r * r3) / 2.0f;
//			result.y = y * r * 1.5f + r;
			result.x = (x + 0.5f) * (r * r3);
			result.y = (y * 1.5f + 1.0f) * r;
		}
		return result;
	}

	// c 𒆐SƂa r ̉~`
	inline BOOL DrawCircle(HDC hdc, POINT c, int r)
	{
		return Ellipse(hdc, c.x - r, c.y - r, c.x + r + 1, c.y + r + 1);
	}
}

//-----------------------------------------------------------------------------
// RXgN^
//-----------------------------------------------------------------------------
Board::Board(int x, int y)
{
	location.x = x;
	location.y = y;

	hBackgroundBrush = CreateSolidBrush(BACKGROUND_COLOR);

	reset();
}

//-----------------------------------------------------------------------------
// fXgN^
//-----------------------------------------------------------------------------
Board::~Board()
{
	DeleteObject(hBackgroundBrush);
}

//-----------------------------------------------------------------------------
// Rs[RXgN^
//-----------------------------------------------------------------------------
Board::Board(const Board& board)
{
	location = board.location;
	hBackgroundBrush = CreateSolidBrush(BACKGROUND_COLOR);
	memcpy(discs, board.discs, sizeof(discs[0][0]) * SIZE * SIZE);
}

//-----------------------------------------------------------------------------
// Zbg
//-----------------------------------------------------------------------------
void Board::reset()
{
	ZeroMemory(discs, sizeof(discs[0][0]) * SIZE * SIZE);

	discs[3][4] = discs[4][3] = Disc::black();
	discs[3][3] = discs[4][4] = Disc::white();

	while (!undoBuffer.empty())
		undoBuffer.pop();

	turn = 0;
}

//-----------------------------------------------------------------------------
// `悷
//-----------------------------------------------------------------------------
void Board::draw(HDC hdc)
{
	POINT prevOrg;

	// _炷
	SetViewportOrgEx(hdc, location.x, location.y, &prevOrg);

	// Zp``
	static const float r = Board::CELL_SIZE / 2.0f; // Zp`̈ӂ̒
	static const float r3 = sqrt(3.0f);             // 3

	SelectObject(hdc, GetStockObject(BLACK_PEN));
	SelectObject(hdc, hBackgroundBrush);
	for (int i = 0; i < SIZE; ++i) {
		for (int j = 0; j < SIZE; ++j) {
			float xo, yo;
			if (j % 2 == 0) {
				xo = i * r * r3;
				yo = j * r * 1.5f;
			} else {
				xo = i * r * r3 + r * (r3 / 2.0f);
				yo = j * r * 1.5f;
			}
			POINT points[] = {
				{ xo, yo + r / 2.0f },
				{ xo, yo + r * 1.5f },
				{ xo + r * (r3 / 2.0f), yo + r * 2.0f },
				{ xo + r * r3, yo + r * 1.5f },
				{ xo + r * r3, yo + r / 2.0f },
				{ xo + r * (r3 / 2.0f), yo },
			};
			Polygon(hdc, points, 6);
		}
	}
	// ΂`
	SelectObject(hdc, GetStockObject(NULL_PEN));
	for (int i = 0; i < SIZE; ++i) {
		for (int j = 0; j < SIZE; ++j) {
			Disc* disc = discs[i][j];
			if (disc == Disc::black()) {
				SelectObject(hdc, GetStockObject(BLACK_BRUSH));
			} else if (disc == Disc::white()) {
				SelectObject(hdc, GetStockObject(WHITE_BRUSH));
			} else {
				SelectObject(hdc, GetStockObject(NULL_BRUSH));
			}
			DrawCircle(hdc, HexPosToScreenPos(i, j), RADIUS);
		}
	}

	// _ɖ߂
	SetViewportOrgEx(hdc, prevOrg.x, prevOrg.y, NULL);
}

//-----------------------------------------------------------------------------
// (x, y) disc u
// uƂłꍇ́Atrue Ԃ
//-----------------------------------------------------------------------------
bool Board::put(int x, int y, Disc* disc)
{
	static const int vx[][6] = {
		{ -1, 1, -1, -1, 0, 0 },
		{ -1, 1,  0,  0, 1, 1 },
	};
	static const int vy[] = { 0, 0, -1, 1, -1, 1 };

	if (!isOn(x, y) || discs[x][y] != NULL)
		return false;

	int count = 0;  // Ԃ΂̐
	UndoData undoData;

	POINT p0 = { x, y };
	undoData.p0 = p0;
	memset(undoData.p, -1, sizeof(POINT) * 8);
	undoData.turn = turn + 1;

	// ܂Ă΂𗠕Ԃ
	for (int i = 0; i < 6; ++i) {
		int px = x, py = y;

		while (true) {
			px += vx[py % 2][i];
			py += vy[i];

			if (!isOn(px, py) || discs[px][py] == NULL)
				break;

			if (discs[px][py] == disc) {
				POINT pt = { px, py };
				undoData.p[i] = pt;

				while (true) {
					px -= vx[(py+1) % 2][i];
					py -= vy[i];

					if (px == x && py == y)
						break;

					discs[px][py] = disc;
					++count;
				}
				break;
			}
		}
	}

	if (count > 0) {
		discs[x][y] = disc;
		++turn;
		undoBuffer.push(undoData);
		return true;
	} else {
		return false;
	}
}

//-----------------------------------------------------------------------------
// (x, y) disc uƂłꍇAtrue Ԃ
//-----------------------------------------------------------------------------
bool Board::canPut(int x, int y, Disc* disc) const
{
	if (!isOn(x, y))
		return false;

	if (discs[x][y] != NULL)
		return false;

	if (getReversibleDiscCount(x, y, disc) == 0)
		return false;

	return true;
}

//-----------------------------------------------------------------------------
// (x, y) disc uƂɁAԂƂ̂ł΂̐擾
//-----------------------------------------------------------------------------
int Board::getReversibleDiscCount(int x, int y, Disc* disc) const
{
	static const int vx[][6] = {
		{ -1, 1, -1, -1,  0,  0 },
		{ -1, 1,  0,  0,  1,  1 },
	};
	static const int vy[] = { 0, 0, -1, 1, -1, 1 };

	if (!isOn(x, y) || discs[x][y] != NULL)
		return 0;

	int sum = 0;

	// ZׂĒׂ
	for (int i = 0; i < 6; ++i) {
		int px = x, py = y;
		int localSum = 0;

		while (true) {
			px += vx[py % 2][i];
			py += vy[i];

			if (!isOn(px, py) || discs[px][py] == NULL)
				break;

			if (discs[px][py] == disc) {
				sum += localSum;
				break;
			}

			++localSum;
		}
	}

	return sum;
}

//-----------------------------------------------------------------------------
// ߂
//-----------------------------------------------------------------------------
void Board::undo()
{
	static const int vx[][6] = {
		{ -1, 1, -1, -1,  0,  0 },
		{ -1, 1,  0,  0,  1,  1 },
	};
	static const int vy[] = { 0, 0, -1, 1, -1, 1 };

	if (undoBuffer.empty())
		return;

	UndoData undoData = undoBuffer.top();
	undoBuffer.pop();

	int x0 = undoData.p0.x, y0 = undoData.p0.y;

	Disc* reversed = Disc::getReversed(discs[x0][y0]);
	discs[x0][y0] = NULL;

	for (int i = 0; i < 6; ++i) {
		int px = x0, py = y0;
		int ex = undoData.p[i].x, ey = undoData.p[i].y;

		if (ex == -1) continue;

		while (true) {
			px += vx[py % 2][i];
			py += vy[i];

			if (px == ex && py == ey)
				break;

			discs[px][py] = reversed;
		}
	}

	turn = undoData.turn - 1;
}
