#include "stdafx.h"
#include "com.h"
#include <vector>
#include <stdlib.h>

static const int infinite = 1000000000;

RandomCom::RandomCom(Board* board, Disc* disc)
	: board(board), disc(disc)
{}


void RandomCom::setDisc(Disc* newDisc)
{
	disc = newDisc;
}

POINT RandomCom::think()
{
	// Kɒu

	std::vector<POINT> points;

	for (int i = 0; i < Board::SIZE; ++i) {
		for (int j = 0; j < Board::SIZE; ++j) {
			if (board->canPut(i, j, disc)) {
				POINT point = { i, j };

				// ꍇ͗D悵Ď
				if ((i == 0 || i == Board::SIZE-1) && (j == 0 || j == Board::SIZE-1))
					return point;

				points.push_back(point);
			}
		}
	}
	Sleep(300);
	return points[rand() % points.size()];
}

///////////////////////////////////////////////////////////////////////////////

MaxCom::MaxCom(Board* board, Disc* disc)
	: board(board), disc(disc)
{}

void MaxCom::setDisc(Disc* newDisc)
{
	disc = newDisc;
}

POINT MaxCom::think()
{
	// ł邾

	std::vector<POINT> points;
	int max = 0;

	for (int i = 0; i < Board::SIZE; ++i) {
		for (int j = 0; j < Board::SIZE; ++j) {
			int count = board->getReversibleDiscCount(i, j, disc);
			POINT point = { i, j };

			// ꍇ͗D悵Ď
			if ((count > 0) && (i == 0 || i == Board::SIZE-1) && (j == 0 || j == Board::SIZE-1))
				return point;

			if (count > max) {
				max = count;
				points.clear();
				points.push_back(point);
			} else if (count == max) {
				points.push_back(point);
			}
		}
	}
	Sleep(300);
	return points[rand() % points.size()];
}

///////////////////////////////////////////////////////////////////////////////

AlphaCom::AlphaCom(Board* board, Disc* disc)
	: board(board), disc(disc)
{}

void AlphaCom::setDisc(Disc* newDisc)
{
	disc = newDisc;
}

// ̎l
POINT AlphaCom::think()
{
	int depth = 7;
	int max = -infinite;
	POINT result = { -1, -1 };

	Disc* turnDisc = board->getTurnDisc();

	if (board->getTurn() >= 45)
		depth = 60;	// Ō܂œǂ݂

	for (int i = 0; i < Board::SIZE; ++i) {
		for (int j = 0; j < Board::SIZE; ++j) {
			if (board->canPut(i, j, turnDisc)) {
				board->put(i, j, turnDisc);
				int score = alphabeta(depth, -infinite, +infinite);
				if (max < score) {
					max = score;
					result.x = i;
					result.y = j;
				}
				board->undo();
			}
		}
	}

	return result;
}

// -@ŒT
int AlphaCom::alphabeta(int depth, int alpha, int beta)
{
	if (depth == 0)
		return eval();
	
	Disc* turnDisc = board->getTurnDisc();

	if (turnDisc == disc) {
		bool pass = true;

		for (int i = 0; i < Board::SIZE; ++i) {
			for (int j = 0; j < Board::SIZE; ++j) {
				if (board->canPut(i, j, turnDisc)) {
					board->put(i, j, turnDisc);
					int score = alphabeta(depth - 1, alpha, beta);
					board->undo();
					if (score >= beta)
						return score;
					else if (score > alpha)
						alpha = score;
					pass = false;
				}
			}
		}

		if (pass) {
			return eval();
		}

		return alpha;
	} else {
		bool pass = true;
		
		for (int i = 0; i < Board::SIZE; ++i) {
			for (int j = 0; j < Board::SIZE; ++j) {
				if (board->canPut(i, j, turnDisc)) {
					board->put(i, j, turnDisc);
					int score = alphabeta(depth - 1, alpha, beta);
					board->undo();
					if (score <= alpha)
						return score;
					else if (score < beta)
						beta = score;
					pass = false;
				}
			}
		}

		if (pass) {
			return eval();
		}

		return beta;
	}
}

// ǖʂ]
int AlphaCom::eval()
{
	static const int points[8][8] = {
		{  30, -12,  0, -1, -1,  0, -12,  30 },
		{ -12,  -1, -3, -3, -3, -3, -15, -12 },
		{  15,  -5,  0, -1, -1,  0,  -5,  15 },
		{  -6,  -5, -1,  1,  1, -1,  -5,  -6 },
		{  15,  -5, -1,  1,  1, -1,  -5,  15 },
		{  -6,  -5,  0, -1, -1,  0,  -5,  -6 },
		{  15, -15, -3, -3, -3, -3,  -1,  15 },
		{  30, -12,  0, -1, -1,  0, -12,  30 },
	};

	int result = 0;
	int turn = board->getTurn();

	for (int i = 0; i < Board::SIZE; ++i) {
		for (int j = 0; j < Board::SIZE; ++j) {
			if (board->get(i, j) == disc) {
				if (turn < 60)
					result += points[i][j];
				else
					result += 1;
			}
		}
	}

	return result;
}
