#include "stdafx.h"
#include "controller.h"
#include "resource.h"
#include <math.h>
#include <assert.h>

namespace {
	Disc* getTurnDisc(int turn) {
		return turn % 2 == 0 ? Disc::black() : Disc::white();
	}

	// XN[WZp`W֕ϊ
	POINT ScreenPosToHexPos(int x, int y)
	{
		static const float r = Board::CELL_SIZE / 2.0f; // Zp`̈ӂ̒
		static const float r3 = sqrt(3.0f);             // 3

		POINT result;

		int h = y % (int)(r * 1.5f);
		int f = y / (int)(r * 1.5f);
		int w = (f % 2) ? (x - r * (r3 / 2.0f)) : (x);

		if (h < r / 2.0f) {
			// (x, y)Zp`̏̕܂͉̕ɂƂ

			w -= (r / 2.0f - h) * r3;

			int t0 = floor(w / floor(r * r3));
			int t1 = w % (int)(r * r3);

			if ((UINT)t1 < h * r3 * 2.0f) {
				result.x = t0;
				result.y = f;
			} else {
				result.x = t0 + (f % 2);
				result.y = f - 1;
			}
		} else {
			// (x, y)Zp`̒`̕ɂƂ

			result.x = floor(w / (r * r3));
			result.y = f;
		}

		return result;
	}

	const int VS_GROUP_FIRST = IDM_VS_HUMAN;
	const int VS_GROUP_LAST  = IDM_VS_COM3;
}

//-----------------------------------------------------------------------------
// RXgN^
//-----------------------------------------------------------------------------
Controller::Controller(HWND hWnd): gameover(false), com(NULL)
{
	board = new Board(10, 10);
	statusDisplay = new StatusDisplay(
		Board::CELL_SIZE / 2 * sqrt(3.0) * (Board::SIZE + 0.5) + 20, 10, board
	);

	RECT rect;
	GetClientRect(hWnd, &rect);
	HDC hdc = GetDC(hWnd);

	// _uobt@Ȍ
	hBufferBitmap = CreateCompatibleBitmap(
		hdc, rect.right-rect.left, rect.bottom-rect.top
	);
	hBufferDC = CreateCompatibleDC(hdc);
	SelectObject(hBufferDC, hBufferBitmap);

	// obt@hԂ
	HBRUSH brush = CreateHatchBrush(HS_DIAGCROSS, RGB(191, 207, 223));
	FillRect(hBufferDC, &rect, brush);
	DeleteObject(brush);

	ReleaseDC(hWnd, hdc);

	srand(GetTickCount());
	onCommand(hWnd, IDM_VS_COM1, NULL);
}

//-----------------------------------------------------------------------------
// fXgN^
//-----------------------------------------------------------------------------
Controller::~Controller()
{
	delete com;
	delete statusDisplay;
	delete board;

	DeleteDC(hBufferDC);
	DeleteObject(hBufferBitmap);
}

//-----------------------------------------------------------------------------
// `揈
//-----------------------------------------------------------------------------
void Controller::onPaint(HWND hWnd)
{
	//  hBufferDC ̂قɕ`悵ĂāǍ hdc ɓ]B
	// 邱ƂŁAʂ̃`hƂłB

	HDC hdc = GetDC(hWnd);

	board->draw(hBufferDC);
	statusDisplay->draw(hBufferDC);

	RECT rect;
	GetClientRect(hWnd, &rect);
	BitBlt(hdc, 0, 0, rect.right-rect.left, rect.bottom-rect.top, hBufferDC, 0, 0, SRCCOPY);

	ReleaseDC(hWnd, hdc);
	ValidateRect(hWnd, NULL);
}

//-----------------------------------------------------------------------------
// NbN
//-----------------------------------------------------------------------------
void Controller::onLButtonUp(HWND hWnd, int x, int y)
{
	if (gameover)
		return;
	POINT pt = ScreenPosToHexPos(x - board->getX(), y - board->getY());
	if (board->put(pt.x, pt.y, getTurnDisc(board->getTurn()))) {
		onPut(hWnd);
	}
	if (gameover)
		onGameOver(hWnd);
	else
		onPaint(hWnd);
}

//-----------------------------------------------------------------------------
// R}h
//-----------------------------------------------------------------------------
void Controller::onCommand(HWND hWnd, int id, HWND hwndCtrl)
{
	const int WHITE_BUTTON = 0;

	switch (id) {
	case IDM_NEWGAME:
		board->reset();
		statusDisplay->update(0, com);
		gameover = false;
		if (com != NULL)
			com->setDisc(Disc::white());
		InvalidateRect(hWnd, NULL, FALSE);
		break;

	case IDM_NEWGAME_WHITE:
		board->reset();
		statusDisplay->update(0, com);
		gameover = false;
		if (com != NULL) {
			com->setDisc(Disc::black());
			onPut(hWnd);
		}
		InvalidateRect(hWnd, NULL, FALSE);
		break;

	case IDM_VS_HUMAN:
		delete com;
		com = NULL;
		CheckMenuRadioItem(
			GetMenu(hWnd), VS_GROUP_FIRST, VS_GROUP_LAST, IDM_VS_HUMAN, MF_BYCOMMAND
		);
		onCommand(hWnd, IDM_NEWGAME, NULL);
		break;

	case IDM_VS_COM1:
		delete com;
		com = new RandomCom(board, Disc::white());
		CheckMenuRadioItem(
			GetMenu(hWnd), VS_GROUP_FIRST, VS_GROUP_LAST, IDM_VS_COM1, MF_BYCOMMAND
		);
		onCommand(hWnd, IDM_NEWGAME, NULL);
		break;

	case IDM_VS_COM2:
		delete com;
		com = new MaxCom(board, Disc::white());
		CheckMenuRadioItem(
			GetMenu(hWnd), VS_GROUP_FIRST, VS_GROUP_LAST, IDM_VS_COM2, MF_BYCOMMAND
		);
		onCommand(hWnd, IDM_NEWGAME, NULL);
		break;

	case IDM_VS_COM3:
		delete com;
		com = new AlphaCom(board, Disc::white());
		CheckMenuRadioItem(
			GetMenu(hWnd), VS_GROUP_FIRST, VS_GROUP_LAST, IDM_VS_COM3, MF_BYCOMMAND
		);
		onCommand(hWnd, IDM_NEWGAME, NULL);
		break;

	case IDM_UNDO: {
		if (gameover)
			break;
		Disc* disc = getTurnDisc(board->getTurn());
		do {
			board->undo();
		} while (disc != getTurnDisc(board->getTurn()));
		statusDisplay->update(board->getTurn(), com, false, NULL);
		onPaint(hWnd);
		break;
	}

	case IDM_EXIT:
		SendMessage(hWnd, WM_DESTROY, 0, 0);
		break;

	case IDM_ABOUT:
		MessageBox(
			hWnd, TEXT("Zpo[V Ver 1.1.0\r\n: 쓇 T"),
			TEXT("About.."), MB_OK
		);
		break;
	}
}

//-----------------------------------------------------------------------------
// ΂uꂽƂ̏
//-----------------------------------------------------------------------------
void Controller::onPut(HWND hWnd)
{
	bool pass = false;

	while (true) {
		if (canPut()) {
			statusDisplay->update(board->getTurn(), com);

			if (pass) {
				onPaint(hWnd);
				MessageBox(hWnd, TEXT("pX"), TEXT("o[V"), MB_OK);
				pass = false;
			}

			if (com && com->getDisc() == getTurnDisc(board->getTurn())) {
				onPaint(hWnd);
				POINT point = com->think();
				if (!board->put(point.x, point.y, com->getDisc()))
					assert(0);
			} else {
				return;
			}
		} else {
			board->pass();
			if (pass) {
				gameover = true;
				return;
			} else {
				pass = true;
			}
		}
	}
}

//-----------------------------------------------------------------------------
// ΂ǂɒuƂłꍇAtrueԂ
//-----------------------------------------------------------------------------
bool Controller::canPut()
{
	Disc* disc = getTurnDisc(board->getTurn());

	for (int i = 0; i < Board::SIZE; ++i) {
		for (int j = 0; j < Board::SIZE; ++j) {
			if (board->canPut(i, j, disc)) {
				return true;
			}
		}
	}
	return false;
}

//-----------------------------------------------------------------------------
// Q[IƂ̏
//-----------------------------------------------------------------------------
void Controller::onGameOver(HWND hWnd)
{
	int blackCount = 0, whiteCount = 0;

	for (int i = 0; i < Board::SIZE; ++i) {
		for (int j = 0; j < Board::SIZE; ++j) {
			Disc* disc = board->get(i, j);

			if (disc == Disc::black())
				++blackCount;
			else if (disc == Disc::white())
				++whiteCount;
		}
	}

	LPCSTR lpText;
	Disc* win;

	if (blackCount == whiteCount) {
		lpText = TEXT("łB");
		win = NULL;
	} else if (blackCount > whiteCount) {
		lpText = TEXT("̏łB");
		win = Disc::black();
	} else {
		lpText = TEXT("̏łB");
		win = Disc::white();
	}

	statusDisplay->update(board->getTurn(), com, gameover, win);
	onPaint(hWnd);
	MessageBox(hWnd, lpText, TEXT("o[V"), MB_OK);
}
